Introduction
============

Crawler Rate Limit allows you to limit requests performed by web crawlers, bots
and spiders. Optionally, it can also rate limit regular (non-bot) traffic.

There are number of crawlers, bots and spiders that are known for excessive
crawling of websites. Such requests increase the server load and often affect
performance of the website for regular visitors. In extreme cases they can even
bring the site down.

This module detects if the request is made by the crawler/bot/spider by
inspecting the UserAgent HTTP header and then limits number of requests crawler
is allowed to perform in the given time interval. After the limit is exceeded
server will respond with HTTP code 429 (Too many requests). Crawler is unblocked
at the end of that time interval and new cycle beings. If crawler exceeds the
rate limit again, it will again be blocked for the duration of the same time
interval.

**Features**
- by default targets only web crawlers, bots and spiders; regular browser
  traffic is not affected
- number of allowed requests in a given time interval (limit) is configurable
- limits bot requests based on the UserAgent string - if the same crawler uses
  multiple IP addresses all those requests will count towards the single limit
- optionally it can also rate limit regular traffic - limiting is based on the
  combination of IP address and User Agent string
- limit for regular traffic is configured independently from the limit for bot
  traffic
- minimal effect on performance
  - rate limiting is performed in a very early phase of request handling
  - uses Redis, APCu or Memcached as rate limiter backend

Requirements
============

Dependencies installed along with the module:
- jaybizzle/crawler-detect - https://github.com/jaybizzle/crawler-detect
- nikolaposa/rate-limit - https://github.com/nikolaposa/rate-limit

Each backend has additional dependencies which must be installed in addition to
the module in order to use that particular backend.

**Redis backend**
- Either Redis PECL extension (https://pecl.php.net/package/redis) or Predis PHP
  package (https://github.com/predis/predis)
- Redis Drupal module - https://www.drupal.org/project/redis

**APCu backend**
- APCu PECL extension - https://pecl.php.net/package/APCu

**Memcached backend**
- Memcached PECL extension - https://pecl.php.net/package/memcached
- Memcache Drupal module - https://www.drupal.org/project/memcache

**Important:** Crawler Rate Limit does not support Mecache PECL extension. Only
Memcached is supported.

Installation and configuration
==============================

First install dependencies for your backend. Follow the steps covering your
backend of choice.

APCu backend
------------

1. Install APCu PECL extension by following instructions provided for your
   operating system.

Redis backend
-------------

1. Install Redis PECL extension by following instructions provided for your
   operating system or install Predis PHP package via composer. It's sufficient
   to install one or the other.

2. Install Drupal Redis module

3. Configure Redis module. Crawler Rate Limit requires only minimum Redis
   configuration (host and port). It's not necessary to configure cache, locking
   or any other Redis backend.

    ```php
    // Example Redis configuration.
    $settings['redis.connection']['host'] = '127.0.0.1';
    $settings['redis.connection']['port'] = '6379';
    ```

Memcache backend
----------------

1. Install Memcached PECL extension by following instructions provided for your
   operating system.

2. Install Drupal Memcache module

3. Configure Memcache module. Crawler Rate Limit requires only minimum Memcache
   configuration (servers and bins). It's not necessary to configure cache,
   locking or any other Memcache backend.

    ```php
    // Example Memcache configuration.
    $settings['memcache']['servers'] = ['127.0.0.1:11211' => 'default'];
    $settings['memcache']['bins'] = ['default' => 'default'];
    $settings['memcache']['key_prefix'] = '';
    ```

Once the backend dependencies are in place, you can install Crawler Rate Limit.

Install Crawler Rate Limit module
---------------------------------

1. Crawler Rate Limit manages its dependencies via composer. Just copying the
   module into the modules folder won't work.

    ```sh
    cd /root/of/your/drupal/project
    composer require drupal/crawler_rate_limit
    drush en crawler_rate_limit
    ```

2. Configure Crawler Rate Limit. Add the following snippet to your
   `settings.php` file.

    ```php
    // Below configuration uses redis backend and will allow crawlers to perform
    // 100 requests within 600 seconds. It will also rate limit regular traffic
    // allowing 300 requests within 600 seconds.

    // Enable or disable rate limiting. Required.
    $settings['crawler_rate_limit.settings']['enabled'] = TRUE;
    // Supported backends: redis, memcached, apcu. Required.
    $settings['crawler_rate_limit.settings']['backend'] = 'redis';
    // Limit for crawlers and bots traffic. Required.
    $settings['crawler_rate_limit.settings']['bot_traffic'] = [
      // Time interval in seconds. Must be whole number greater than zero.
      'interval' => 600,
      // Number of requests allowed in the given time interval. Must be whole
      // number greater than zero.
      'requests' => 100,
    ];
    // Limit for regular website traffic. Optional. If this setting is not
    // provided, regular traffic won't be rate limited. Set both values to 0 in
    // order to disable limiting of regular traffic.
    $settings['crawler_rate_limit.settings']['regular_traffic'] = [
      'interval' => 600,
      'requests' => 300,
    ];
    ```

    - Make sure to adjust the value for `backend` to match the one you decided
      to use.
    - Make sure to adjust the values for `interval` and `requests` according
      to the capabilities of your server and level of the traffic your website
      receives. You may need to review your server logs in order to determine
      optimal values for your server.

3. Visit Status report page (/admin/reports/status) and confirm that Crawler
   Rate Limit is enabled and it doesn't report any warnings or errors.

Maintainers
===========

- Vojislav Jovanović (vaish) - https://www.drupal.org/u/vaish


Supporting organizations:
- CED Greentech - https://www.drupal.org/ced-greentech
