<?php

namespace Drupal\crawler_rate_limit;

use Drupal\Core\Site\Settings;
use Drupal\Core\StreamWrapper\PublicStream;
use Jaybizzle\CrawlerDetect\CrawlerDetect;
use RateLimit\Exception\LimitExceeded;
use RateLimit\Rate;
use Symfony\Component\HttpFoundation\Request;

/**
 * Rate Limit Manager.
 *
 * @package Drupal\crawler_rate_limit
 */
class RateLimitManager implements RateLimitManagerInterface {

  /**
   * Prefix for the backend (rate limiter) keys.
   */
  const CRAWLER_RATE_LIMIT_KEY_PREFIX = 'crawler_rate_limit:';

  /**
   * Rate Limit backend.
   *
   * @var string
   */
  private string $backend;

  /**
   * Flag indicating whether rate limiter is enabled or not.
   *
   * @var bool
   */
  private bool $enabled;

  /**
   * Flag indicating whether to also limit regular website traffic (non-bots).
   *
   * @var bool
   */
  private bool $limitRegular;

  /**
   * Rate Limit backend factory.
   *
   * @var \Drupal\crawler_rate_limit\RateLimitBackendFactory
   */
  private RateLimitBackendFactory $factory;

  /**
   * Interval in seconds for regular traffic.
   *
   * @var int
   */
  private int $intervalRegular;

  /**
   * Number of requests that is allowed in the interval for regular traffic.
   *
   * @var int
   */
  private int $requestsRegular;

  /**
   * Interval in seconds for bot traffic.
   *
   * @var int
   */
  private int $intervalBots;

  /**
   * Number of requests that is allowed in the interval for bot traffic.
   *
   * @var int
   */
  private int $requestsBots;

  /**
   * Value to use in the Retry-After HTTP header if the request is blocked.
   *
   * @var int
   */
  private int $retryAfter;

  /**
   * CrawlerDetect. Detects bots/crawlers/spiders via the user agent.
   *
   * @var \Jaybizzle\CrawlerDetect\CrawlerDetect
   */
  private CrawlerDetect $crawlerDetect;

  /**
   * Class constructor.
   *
   * @param \Jaybizzle\CrawlerDetect\CrawlerDetect $crawlerDetect
   *   CrawlerDetect object.
   * @param \Drupal\crawler_rate_limit\RateLimitBackendFactory $factory
   *   Rate Limit backend factory.
   */
  public function __construct(CrawlerDetect $crawlerDetect, RateLimitBackendFactory $factory) {
    $this->crawlerDetect = $crawlerDetect;
    $this->factory = $factory;

    $settings = self::getSettings();
    $this->backend = $settings['backend'];
    $this->enabled = $settings['enabled'];
    $this->intervalRegular = $settings['regular_traffic']['interval'];
    $this->requestsRegular = $settings['regular_traffic']['requests'];
    $this->intervalBots = $settings['bot_traffic']['interval'];
    $this->requestsBots = $settings['bot_traffic']['requests'];
    $this->limitRegular = $settings['limit_regular'];
    // Set retryAfter to higher interval value until we figure out which type of
    // request we are handling in limit() method.
    $this->retryAfter = $this->intervalRegular > $this->intervalBots ? $this->intervalRegular : $this->intervalBots;
  }

  /**
   * {@inheritdoc}
   */
  public function limit(Request $request) {
    if (!$this->isEnabled()) {
      return FALSE;
    }

    if ($this->crawlerDetect->isCrawler()) {
      $identifier = $this->crawlerDetect->getMatches();
      $requests = $this->requestsBots;
      $interval = $this->intervalBots;
    }
    elseif ($this->limitRegular) {
      // Some page requests trigger additional request which should not be
      // counted towards the limit. Such requests are ignored.
      // Image module: generate image derivatives of publicly available files.
      // - path: /[file_public_path]/styles/ (e.g. /sites/default/files/styles)
      // Image module: generate image derivatives of private files.
      // - path: /system/files/styles/
      // History module: mark a node as read by the current user.
      // - path: /history/{node}/read (e.g. /history/18/read)
      // Contextual Links module: render contextual links.
      // - path: /contextual/render
      // Media module: render an oEmbed resource.
      // - path: /media/oembed.
      $styles_public = PublicStream::basePath() . '/styles/';
      $pattern = "@($styles_public)|(/system/files/styles/)|(/history/\d+/read)|(/contextual/render)|(/media/oembed)@";
      if (preg_match($pattern, $request->getRequestUri())) {
        return FALSE;
      }

      // As identifier we use combination of IP address and User Agent string.
      // @todo Change hashing algorithm to "xxh3" once the support for PHP 7.4
      // is removed.
      $identifier = hash('crc32c', $request->getClientIp() . $request->headers->get('user-agent'));
      $requests = $this->requestsRegular;
      $interval = $this->intervalRegular;
    }
    else {
      return FALSE;
    }
    $this->retryAfter = $interval;

    try {
      $rate = Rate::custom($requests, $interval);
      $rate_limiter = $this->factory->get($this->backend, $rate, self::CRAWLER_RATE_LIMIT_KEY_PREFIX);
    }
    catch (\Exception $exception) {
      // Prevent missing dependencies or invalid settings to cause fatal errors.
      // All these will be reported on the Status report page.
      return FALSE;
    }

    try {
      $rate_limiter->limit($identifier);
    }
    catch (LimitExceeded $exception) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function isEnabled() {
    return $this->enabled;
  }

  /**
   * {@inheritdoc}
   */
  public function retryAfter() {
    return $this->retryAfter;
  }

  /**
   * Converts value to integer if possible. Otherwise returns 0.
   *
   * @param mixed $value
   *   Value to be converted.
   *
   * @return int
   *   Value converted to integer.
   */
  public static function toIntOrZero($value) {
    if (is_int($value)) {
      return $value;
    }

    if (ctype_digit($value)) {
      return (int) $value;
    }

    return 0;
  }

  /**
   * Returns Crawler Rate Limit settings.
   *
   * @return array
   *   Crawler Rate Limit settings.
   */
  public static function getSettings() {
    $settings = Settings::get('crawler_rate_limit.settings', []);
    // Provide default values that will allow Drupal to run without throwing any
    // Exceptions while keeping the rate limiter disabled.
    $settings += [
      'backend' => 'undefined',
      'enabled' => FALSE,
      'bot_traffic' => [
        'interval' => 0,
        'requests' => 0,
      ],
      'regular_traffic' => [
        'interval' => 0,
        'requests' => 0,
      ],
      'limit_regular' => FALSE,
    ];

    // If user-provided values for interval and requests can't be converted to
    // integer avoid throwing the exception in constructor by setting the values
    // to 0. These are still invalid values for Crawler Rate Limit but will
    // allow Drupal to bootstrap and user will be able to see the error reported
    // on the "Status report" page.
    $settings['bot_traffic']['interval'] = self::toIntOrZero($settings['bot_traffic']['interval']);
    $settings['bot_traffic']['requests'] = self::toIntOrZero($settings['bot_traffic']['requests']);
    $settings['regular_traffic']['interval'] = self::toIntOrZero($settings['regular_traffic']['interval']);
    $settings['regular_traffic']['requests'] = self::toIntOrZero($settings['regular_traffic']['requests']);

    if (empty($settings['backend'])) {
      $settings['backend'] = t('undefined');
    }

    // Disable the limiter if invalid settings have been detected.
    if (
      !in_array($settings['backend'], ['redis', 'memcached', 'apcu']) ||
      $settings['bot_traffic']['interval'] <= 0 ||
      $settings['bot_traffic']['requests'] <= 0
    ) {
      $settings['enabled'] = FALSE;
    }

    // Determine if regular traffic should be limited or not.
    if (
      $settings['regular_traffic']['interval'] > 0 &&
      $settings['regular_traffic']['requests'] > 0
    ) {
      $settings['limit_regular'] = TRUE;
    }

    return $settings;
  }

}
